-- ForthEC code optimiser
-- /Mic, 2004


include parser.e


-- x86 registers
constant regs32 = {"eax","ebx","ecx","edx","esi","edi","ebp","esp"}
constant regs16 = {"ax","bx","cx","dx","si","di","bp","sp"}
constant regs8 = {"al","ah","bl","bh","cl","ch","dl","dh"}

-- x86 condition codes ("nz" is left out because of how these are used in the patterns below)
constant conds = {"g","ge","l","le","z"}



-- Integer instruction patterns to scan for
--
-- Examples of rules:
--
-- {
--	{
--	"mov /reg32/,/reg32/",
--	"add [esp],$1"
--	},
--	{{"add [esp],$2"}}
-- }
--
-- This would match e.g.
-- 	
--	mov ecx,eax
--	add [esp],ecx
--
-- and replace it with
--
--	add [esp],eax
--
constant intpatterns = {
{
	{
	"push /reg32/",
	"pop $1"
	},
	{}
},
{
	{
	"mov /reg32/,/reg32/",
	"add [esp],$1"
	},
	{{"add [esp],$2"}}
},
{
	{
	"mov /reg32/,[/reg32/]",
	"push $1"
	},
	{{"push dword ptr [$2]"}}
},
{
	{
	"push /reg32/",
	{"mov /reg32/,/const/","mov /reg32/,/var/","mov /reg32/,/imm/"},
	"pop /reg32/",
	"cmp $4,$2"
	},
	{{"cmp $1,$3"}}
},
{
	{
	"sub esp,4",
	"mov dword ptr [esp],/imm/",
	"pop eax",
	"mov /const/,eax"
	},
	{{"mov $2,$1"}}
},
{
	{
	"mov eax,/imm/",
	"mov /const/,eax"
	},
	{{"mov $2,$1"}}
},
{
	{
	"push /const/",
	"add esp,4"
	},
	{}
},
{
	{
	"mov eax,ipstackptr",
	"mov ecx,[eax]",
	"add ipstackptr,4",
	"push ecx",
	"ret"
	},
	{1,{"add ipstackptr,4"},{"jmp dword ptr [eax]"}}
},
{
	{
	"push dword ptr [esp]",
	"mov eax,/imm/",
	"pop /reg32/",
	"cmp $2,eax"
	},
	{{"cmp dword ptr [esp],$1"}}
},
{
	{
	"mov eax,[esp]",
	"add /reg32/,eax"
	},
	{{"add $1,[esp]"}}
},
{
	{
	"push eax",
	"mov ebx,dictptr",
	"pop eax",
	"add dictptr,4",
	"mov [ebx],eax"
	},
	{2,4,5}
},
{
	{
	"sub esp,4",
	"mov dword ptr [esp],/imm/",
	"mov ebx,dictptr",
	"pop eax",
	"add dictptr,4",
	"mov [ebx],eax"
	},
	{3,5,{"mov dword ptr [ebx],$1"}}
},
{
	{
	"sub esp,/imm/",
	"mov dword ptr [esp],/imm/",
	"sub esp,/imm/",
	"mov dword ptr [esp],/imm/"
	},
	{{"mov dword ptr [esp-$1],$2"},{"mov dword ptr [esp-%($1+$3)],$4"},{"sub esp,%($1+$3)"}}
},
{
	{
	"mov dword ptr [esp-/imm/],/imm/",
	"mov dword ptr [esp-/imm/],/imm/",
	"sub esp,/imm/",
	"mov dword ptr [esp-/imm/],/imm/",
	"mov dword ptr [esp-/imm/],/imm/",
	"sub esp,/imm/"
	},
	{1,2,{"mov dword ptr [esp-%($5+$6)],$7"},{"mov dword ptr [esp-%($5+$8)],$9"},{"sub esp,%($5+$10)"}}
},
{
	{
	"lea esp,[esp+/imm/]",
	"lea esp,[esp+/imm/]"
	},
	{{"lea esp,[esp+%($1+$2)]"}}
},
{
	{
	"mov eax,[esp+/imm/]",
	"push eax"
	},
	{{"push dword ptr [esp+$1]"}}
},
{
	{
	"set/cond/ cl",
	"neg ecx",
	"mov [esp],ecx",
	"pop eax",
	"cmp eax,-1",
	"jnz /label/"
	},
	{4,{"jn$1 $2"}}
},
{
	{
	"add [esp],/reg32/",
	"mov eax,[esp]",
	"mov ecx,[esp+4]",
	{"mov [eax],/reg8/"}
	},
	{2,3,{"mov [eax+$1],$2"}}
},
{
	{
	"push /reg32/",
	{"push /reg32/","push /var/","push /const/"},
	"mov eax,[esp]",
	"mov ecx,[esp+4]",
	{"mov [eax],/reg32/","mov [eax],/reg16/","mov [eax],/reg8/"},
	"add esp,8"
	},
	{{"mov ecx,$1"},{"mov eax,$2"},{"mov [eax],$3"}}
},
{
	{
	"sub esp,4",
	"mov dword ptr [esp],/imm/",
	{"push /reg32/","push /var/","push /const/"},
	"mov eax,[esp]",
	"mov ecx,[esp+4]",
	{"mov [eax+/reg32/],/reg32/", "mov [eax+/reg32/],/reg16/", "mov [eax+/reg32/],/reg8/"},
	"add esp,8"
	},
	{{"mov ecx,$1"},{"mov eax,$2"},{"mov [eax+$3],$4"}}
},
{
	{
	"sub esp,4",
	"mov dword ptr [esp],/imm/",
	{"push /reg32/","push /var/","push /const/"},
	"mov eax,[esp]",
	"mov ecx,[esp+4]",
	{"mov [eax],/reg32/", "mov [eax],/reg16/", "mov [eax],/reg8/"},
	"add esp,8"
	},
	{{"mov ecx,$1"},{"mov eax,$2"},{"mov [eax],$3"}}
},
{
	{
	"mov eax,/var/",
	"push dword ptr [eax]",
	"mov eax,/var/",
	"push dword ptr [eax]",
	"mov eax,/const/",
	"add [esp],eax",
	"mov eax,[esp]",
	"mov ecx,[esp+4]",
	"mov [eax],ecx",
	"add esp,8"
	},
	{{"mov ecx,$2"},1,{"mov ecx,[ecx]"},{"mov eax,[eax]"},{"add ecx,$3"},{"mov [ecx],eax"}}
},
{
	{
	"push dword ptr [esp]",
	"pop /reg32/"
	},
	{{"mov $1,[esp]"}}
},
{
	{
	"mov /reg32/,[esp]",
	"add /reg32/,$1"
	},
	{{"add $2,[esp]"}}
},
{
	{
	"mov ecx,eax",
	"mov eax,/var/",
	"mov [eax],ecx"
	},
	{{"mov ecx,$1"},{"mov [ecx],eax"}}
},
{
	{
	"mov eax,/imm/",
	"mov ecx,[esp]",
	"mov [ecx],eax",
	"add dword ptr [esp],4"
	},
	{{"mov eax,[esp]"},{"mov dword ptr [eax],$1"},4}
},
{	
	{
	"push dword ptr [eax]",
	{"mov eax,/const/","mov eax,/var/"},
	"add [esp],eax",
	"pop eax"
	},
	{{"mov eax,[eax]"},{"add eax,$1"}}
},
{	
	{
	"sub esp,/imm/",
	"sub esp,/imm/"
	},
	{{"sub esp,%($1+$2)"}}
},
{
	{
	"mov eax,/imm/",
	"add [/reg32/],eax"
	},
	{{"add dword ptr [$2],$1"}}
},
{
	{
	"sub ipstackptr,4",
	"pop ecx",
	"mov eax,ipstackptr",
	"mov [eax],ecx"
	},
	{1,3,{"pop dword ptr [eax]"}}
},
{	
	{
	"sub ipstackptr,4",
	"mov eax,ipstackptr",
	"pop dword ptr [eax]",
	"sub ipstackptr,4",
	"mov eax,ipstackptr",
	"mov [eax],ebp",
	"mov ebp,esp"
	},
	{2,{"pop dword ptr [eax-4]"},{"sub ipstackptr,8"},{"mov [eax-8],ebp"},{"mov ebp,esp"}}
},
{	
	{
	{"push /var/","push /ind/"},
	"mov eax,loopstackptr",
	"mov eax,dword ptr [eax-4]",
	"add [esp],eax",
	"pop /reg32/"
	},
	{{"mov $2,$1"},{"mov edx,loopstackptr"},{"add $2,dword ptr [edx-4]"}}
},
{
	{
	"fld /qlit/",
	"sub esp,4",
	"fstp dword ptr [esp]",
	"fld /qlit/",
	"sub esp,4",
	"fstp dword ptr [esp]"
	},
	{1,4,{"sub esp,8"},{"fstp dword ptr [esp]"},{"fstp dword ptr [esp+4]"}}
},
{
	{
	"mov /reg32/,$1"
	},
	{}
}
}


-- Floating point instruction patterns to scan for
constant fppatterns = {
{
	{
	"mov eax,fpstackptr",		-- The pattern
	"fld qword ptr [lit_",
	"add fpstackptr,8",
	"fstp qword ptr [eax]",
	"mov eax,fpstackptr",
	"sub esp,4",
	"fld qword ptr [eax-8]",
	"sub fpstackptr,8",
	"fstp dword ptr [esp]"
	},
	{1,2,6,9}			-- Which instructions to keep	
},
{
	{
	"mov eax,fpstackptr",
	"fld qword ptr [eax-8]",
	{"fadd qword ptr [eax-16]","fsub qword ptr [eax-16]","fmul qword ptr [eax-16]","fdiv qword ptr [eax-16]"},
	"sub fpstackptr,8",
	"fstp qword ptr [eax-16]",
	"mov eax,fpstackptr",
	"fld qword ptr [eax-8]",
	{"fadd qword ptr [eax-16]","fsub qword ptr [eax-16]","fmul qword ptr [eax-16]","fdiv qword ptr [eax-16]"},
	"sub fpstackptr,8",
	"fstp qword ptr [eax-16]"
	},
	{1,2,3,{"sub eax,8"},{"sub fpstackptr,16"},8,10}
},
{
	{
	"mov eax,dword ptr [var_",
	"mov ebx,fpstackptr",
	"fld qword ptr [eax]",
	"add fpstackptr,8",
	"fstp qword ptr [ebx]",
	"mov eax,dword ptr [var_",
	"mov ebx,fpstackptr",
	"fld qword ptr [eax]",
	"add fpstackptr,8",
	"fstp qword ptr [ebx]",
	"mov eax,fpstackptr",
	"fld qword ptr [eax-8]",
	{"fadd qword ptr [eax-16]","fsub qword ptr [eax-16]","fmul qword ptr [eax-16]","fdiv qword ptr [eax-16]"},
	"sub fpstackptr,8",
	"fstp qword ptr [eax-16]"
	},
	{6,2,3,1,8,{"add fpstackptr,8"},{13,1,4," st(0),st(1)"},{"fstp qword ptr [ebx]"}}
},
{
	{
	"mov eax,fpstackptr",
	"fld qword ptr [lit",
	"sub esp,4",
	"fstp dword ptr [esp]"
	},
	{2,3,4}
}
}



global integer
	linesRemoved,
	constRemoved,
	optLevel

global sequence
	patvars,
	constlist
	
linesRemoved = 0
constRemoved = 0
constlist = {}


-- Is this a pop instruction ?
function is_pop(sequence s)
	if length(s)>3 then
		if equal(s[1..3],"pop") then
			return 1
		end if
	end if
	return 0
end function


-- Is this a push instruction ?
function is_push(sequence s)
	if length(s)>4 then
		if equal(s[1..4],"push") then
			return 1
		elsif length(s)>19 then
			if equal(s[1..19],"mov dword ptr [esp]") then
				return 2
			end if
		end if
	end if
	return 0
end function


function get_reg(sequence s)
	integer p
	
	p = find(',',s)
	if p then
		return s[p+1..length(s)]
	else
		return s[find(' ',s)+1..length(s)]
	end if
end function


function get_reg1(sequence s)
	integer p
	
	p = find(',',s)
	if p then
		return s[find(' ',s)+1..p-1]
	else
		return s[find(' ',s)+1..length(s)]
	end if
end function


function get_imm(sequence s)
	return s[find(',',s)+1..length(s)]
end function


-- safe_compare(code,pattern)
function safe_compare(sequence s1,sequence s2)
	if length(s1)>=length(s2) then
		return equal(s1[1..length(s2)],s2)
	end if
	return 0
end function


function compare_patterns(sequence s1,sequence s2)
	integer eql,p1,p2,m,n,o
	sequence s3,s4
	
	eql = 1
	p1 = 1
	p2 = 1
	m = length(patvars)
	o = 0
	
	
	while eql do
		if p1>length(s1) or
		   p2>length(s2) then
			exit
		end if
		if s2[p2] = '/' then
			s3 = {}
			p2 += 1
			while s2[p2] != '/' do
				s3 &= s2[p2]
				p2 += 1
			end while
			p2 += 1
			
			s4 = {}
			while p1<=length(s1) do
				if s1[p1]=',' or
				   s1[p1]=']' then
					exit
				end if
				s4 &= s1[p1]
				p1 += 1
			end while
			
			if equal(s3,"reg32") then
				if find(s4,regs32) then
					patvars = append(patvars,s4)
				else
					eql = 0
					exit
				end if
			elsif equal(s3,"reg16") then
				if find(s4,regs16) then
					patvars = append(patvars,s4)
				else
					eql = 0
					exit
				end if
			elsif equal(s3,"reg8") then
				if find(s4,regs8) then
					patvars = append(patvars,s4)
				else
					eql = 0
					exit
				end if
				o = 1
			elsif equal(s3,"const") then
				s4 &= ']'
				eql = 0
				if find('_',s4) and length(s4)>=17 then
					if equal("dword ptr [const_",s4[1..17]) then
						eql = 1
						p1 += 1
						patvars = append(patvars,s4)
					end if
				end if
			elsif equal(s3,"var") then
				s4 &= ']'
				eql = 0
				if find('_',s4) then
					if equal("dword ptr [var_",s4[1..15]) then
						eql = 1
						p1 += 1
						patvars = append(patvars,s4)
					end if
				end if
			elsif equal(s3,"ind") then
				s4 &= ']'
				eql = 0
				if length(s4)>12 then
					if equal("dword ptr [e",s4[1..12]) then
						eql = 1
						p1 += 1
						patvars = append(patvars,s4)
					end if
				end if
			elsif equal(s3,"qlit") then
				s4 &= ']'
				eql = 0
				if find('_',s4) then
					if equal("qword ptr [lit_",s4[1..15]) then
						eql = 1
						p1 += 1
						patvars = append(patvars,s4)
					end if
				end if
			elsif equal(s3,"imm") then
				o = 0
				if length(s4)>6 then
					if equal(s4[1..6],"offset") then
						patvars = append(patvars,s4)
						o = 1
					end if
				end if
				if o=0 then
					s3 = value(s4)
					if s3[1]=GET_SUCCESS and integer(s3[2]) then
						patvars = append(patvars,s4)
					else
						eql = 0
					end if
				end if
			elsif equal(s3,"cond") then
				if find(' ',s4) then
					s4 = s4[1..find(' ',s4)-1]
				end if
				--puts(1,s4&"\n")
				if find(s4,conds) then
					patvars = append(patvars,s4)
				else
					eql = 0
				end if
			elsif equal(s3,"label") then
				if s4[1]='@' then
					patvars = append(patvars,s4)
				else
					eql = 0
				end if
			end if
			
		elsif s2[p2]='$' then
			n = 0
			p2 += 1
			while p2<=length(s2) do
				if s2[p2]>='0' and s2[p2]<='9' then
					n = n*10 + s2[p2]-'0'
				else
					exit
				end if
				p2 += 1
			end while
			s3 = patvars[n]

			s4 = {}
			while p1<=length(s1) do
				if s1[p1]=',' then
					exit
				end if
				s4 &= s1[p1]
				p1 += 1
			end while
			
			if not equal(s3,s4) then
				eql = 0
			end if
				
		else
			if s1[p1]!=s2[p2] then
				eql = 0
			end if
			p1 += 1
			p2 += 1
		end if
	end while
	
	if p1<length(s1) or p2<length(s2) then
		eql = 0
	end if
	
	if not eql then
		if m=0 then
			patvars = {}
		else
			patvars = patvars[1..m]
		end if
	end if
	
	return eql
end function



function pattern_append(sequence s1,sequence s2)
	integer p2,n,oper,expval
	sequence s3,s4,ops

	p2 = 1
	s3 = {}
	while 1 do
		if p2>length(s2) then
			exit
		end if
		
		if s2[p2]='$' then
			n = 0
			p2 += 1
			while p2<=length(s2) do
				if s2[p2]>='0' and s2[p2]<='9' then
					n = n*10 + s2[p2]-'0'
				else
					exit
				end if
				p2 += 1
			end while
			s3 &= patvars[n]
		elsif s2[p2]='%' then
			p2 += 2
			expval = 0
			ops = {}
			oper = 0
			while s2[p2] != ')' do
				if s2[p2] = '$' then
					n = 0
					p2 += 1
					while p2<=length(s2) do
						if s2[p2]>='0' and s2[p2]<='9' then
							n = n*10 + s2[p2]-'0'
						else
							exit
						end if
						p2 += 1
					end while
					s4 = value(patvars[n])
					if oper=0 then
						expval = s4[2]
					elsif oper='+' then
						expval += s4[2]
					elsif oper='-' then
						expval -= s4[2]
					elsif oper='*' then
						expval *= s4[2]
					end if

				elsif s2[p2] = '+' then
					oper = s2[p2]
					p2 += 1
				elsif s2[p2] = '+' then
					oper = s2[p2]
					p2 += 1
				elsif s2[p2] = '+' then
					oper = s2[p2]
					p2 += 1
				else
					p2 += 1
				end if
			end while
			p2 += 1
			s3 &= sprintf("%d",expval)
					
		else
			s3 &= s2[p2]
			p2 += 1
		end if
	end while
	
	return append(s1,s3)
end function




global function optimise(sequence subject,integer remConst)
	integer p,n,m,clean,improvement,times
	sequence s,r,pat

	if remConst then
		-- Remove unused constants
		for i=1 to length(constants[2]) do
			if constants[2][i][2]=0 then
				subject = subject[1..constants[2][i][3]-1] & {"add esp,4"} & subject[constants[2][i][3]+2..length(subject)]
				constRemoved += 1
				for j=i+1 to length(constants[2]) do
					if constants[2][j][3]>constants[2][i][3] then
						constants[2][j][3] -= 1
					end if
				end for
			end if
		end for
	end if

	
	if optLevel>=2 then
	
	-- Collapse combinations like push reg/pop reg	
	p = 2
	while p<=length(subject) do
		if is_pop(subject[p]) then
			n = is_push(subject[p-1])
			if n=1 then
				r = get_reg(subject[p])
				if find(r,regs32) then
					if not equal(r,get_reg(subject[p-1])) then
						s = "mov "&r&","&get_reg(subject[p-1])
						if p-2>0 then
							if p+1<=length(subject) then
								subject = subject[1..p-2] & {s} & subject[p+1..length(subject)]
							else
								subject = subject[1..p-2] & {s}
							end if
							linesRemoved += 1
						end if
					end if
				end if
			elsif n=2 then
				r = get_reg(subject[p])
				if find(r,regs32) then
					s = "mov "&get_reg(subject[p])&","&get_imm(subject[p-1])
					--puts(1,"Collapsing "&subject[p-2]&" and "&subject[p-1]&" and "&subject[p]&" into "&s&"\n")
					if p-3>0 then
						if p+1<=length(subject) then
							subject = subject[1..p-3] & {s} & subject[p+1..length(subject)]
						else
							subject = subject[1..p-3] & {s}
						end if
						linesRemoved += 2
					end if
				end if
			end if
		end if
		p += 1
	end while
	
	-- Remove combinations like sub esp,4/mov dword ptr [esp],x/add esp,4
	clean = 0
	while not clean do	-- Repeat until the entire code sequence was scanned without finding anything to remove
		p = 1
		clean = 1
		while p<=length(subject)-2 do
			if length(subject[p])=9 then
				if equal(subject[p],"sub esp,4") then
					if length(subject[p+2])=9 then
						if equal(subject[p+2],"add esp,4") then
							if length(subject[p+1])>19 then
								if equal(subject[p+1][1..19],"mov dword ptr [esp]") then
									if p+3<=length(subject) then
										subject = subject[1..p-1] & subject[p+3..length(subject)]
										linesRemoved += 3
										clean = 0
									end if
								end if
							end if
						end if
					end if
				end if
			end if
			p += 1
		end while
	end while

	-- Collapse combination like mov eax,imm/mov const_x,eax
	p = 2
	while p<=length(subject) do
		if length(subject[p])>21 then
			if equal(subject[p][1..21],"mov dword ptr [const_") then
				if equal(get_reg(subject[p]),"eax") then
					if length(subject[p-1])>7 then
						if equal(subject[p-1][1..7],"mov eax") then
							s = get_imm(subject[p-1])
							r = value(s)
							if r[1] = GET_SUCCESS then	-- Make sure it's a number
								if p>2 and p<length(subject) then
									subject = subject[1..p-2] &
									          --{"mov dword ptr ["&subject[p][5..find(',',subject[p])-1]&"],"&s} &
									          {"mov "&subject[p][5..find(',',subject[p])-1]&","&s} &
									          subject[p+1..length(subject)]
									linesRemoved += 1
								end if
							end if
						end if
					end if
				end if
			end if
		end if
		p += 1
	end while
	
	-- Remove combinations like add esp,4/sub esp,4
	p = 1
	while p<length(subject) do
		if equal(subject[p],"add esp,4") then
			if equal(subject[p+1],"sub esp,4") then
				if p>1 and p+1<length(subject) then
					subject = subject[1..p-1] & subject[p+2..length(subject)]
					linesRemoved += 2
				end if
			end if
		end if
		p += 1
	end while
	
	end if
	

	if optLevel>=3 then
	
		-- Replace common floating point instruction patterns by more compact forms
		for pass=1 to 2 do
			improvement = linesRemoved
			
			p = 1
			while p<=length(subject) do
				for i=1 to length(fppatterns) do
					pat = fppatterns[i]
					if p+length(pat[1])<=length(subject)+1 then
						if safe_compare(subject[p],pat[1][1]) then
							n = 1
							for j=2 to length(pat[1]) do
								if sequence(pat[1][j][1]) then
									m = 0
									for k=1 to length(pat[1][j]) do
										if safe_compare(subject[p+j-1],pat[1][j][k]) then
											m = k
											exit
										end if
									end for
									if not m then
										n = 0
										exit
									end if
								else
									if not safe_compare(subject[p+j-1],pat[1][j]) then
										n = 0
										exit
									end if
								end if
							end for
							if n then
								s = {}
								for j=1 to length(pat[2]) do
									if sequence(pat[2][j]) then
										if sequence(pat[2][j][1]) then
											s = append(s,pat[2][j][1])
										else
											s = append(s,subject[p+pat[2][j][1]-1][pat[2][j][2]..pat[2][j][3]]&pat[2][j][4])
										end if
									else
										s = append(s,subject[p+pat[2][j]-1])
									end if
								end for
								subject = subject[1..p-1] & s & subject[p+length(pat[1])..length(subject)]
								linesRemoved += length(pat[1])-length(pat[2])
								exit
							end if
						end if
					end if
				end for
				p += 1
			end while
			
			improvement -= linesRemoved
			if not improvement then
				exit
			end if
		end for
	end if

	if optLevel >= 5 then
		-- Run up to 10 passes
		times = 10
		while times>0 do
			improvement = linesRemoved
			
			-- Replace common integer instruction patterns by more compact forms
			p = 1
			while p<=length(subject) do
				for i=1 to length(intpatterns) do
					pat = intpatterns[i]
					if p+length(pat[1])<=length(subject)+1 then
						patvars = {}
						--if compare_patterns(subject[p],pat[1][1]) then
							n = 1
							for j=1 to length(pat[1]) do

								if sequence(pat[1][j][1]) then
									m = 0
									for k=1 to length(pat[1][j]) do
										if compare_patterns(subject[p+j-1],pat[1][j][k]) then
											m = k
											exit
										end if
									end for
									if not m then
										n = 0
										exit
									end if
								else
									if not compare_patterns(subject[p+j-1],pat[1][j]) then
										n = 0
										exit
									end if
								end if
							end for

							-- Was a matching pattern found ?
							if n then
								s = {}
								for j=1 to length(pat[2]) do
									if sequence(pat[2][j]) then
										if sequence(pat[2][j][1]) then
											s = pattern_append(s,pat[2][j][1])
										else
											s = pattern_append(s,subject[p+pat[2][j][1]-1][pat[2][j][2]..pat[2][j][3]]&pat[2][j][4])
										end if
									else
										s = pattern_append(s,subject[p+pat[2][j]-1])
									end if
								end for
								subject = subject[1..p-1] & s & subject[p+length(pat[1])..length(subject)]
								linesRemoved += length(pat[1])-length(pat[2])
								exit
							end if
						--end if
					end if
				end for
				p += 1
			end while

			
			-- Stop when there's nothing more to optimise
			improvement -= linesRemoved
			if not improvement then
				exit
			end if
			times -= 1
		end while
	end if
	
	if optLevel >= 5 then
		if remConst then
			constlist = {}
			for i=1 to length(subject) do
				patvars = {}
				if compare_patterns(subject[i],"mov /const/,/imm/") then
					constlist = append(constlist,patvars&i&1)
				end if
			end for
		end if

		patvars = {}
		for i=1 to length(constlist) do
			patvars = constlist[i][1..2]
			for j=1 to length(subject) do
				if equal(subject[j],"push "&patvars[1]) then
					subject[j] = "pushdw "&patvars[2]
					constlist[i][4] = 0
				elsif compare_patterns(subject[j][4..length(subject[j])]," /reg32/,$1") then
					subject[j] = subject[j][1..find(',',subject[j])-1]&","&patvars[2]
					patvars = patvars[1..2]
					constlist[i][4] = 0
				end if
			end for
		end for

		p=2
		while 1 do
			if p+2>length(subject) then
				exit
			end if
			if equal(subject[p],"sub esp,4") then
				patvars = {}
				if compare_patterns(subject[p+1],"mov dword ptr [esp],/imm/") then
					subject = subject[1..p-1] & {"pushdw "&patvars[1]} & subject[p+2..length(subject)]
					linesRemoved += 1
				end if
			end if
			p += 1
		end while
		
		--for i=1 to length(subject) do
		--	if equal(subject[i],"add dword ptr [esp],1") then
		--		subject[i] = "inc dword ptr [esp]"
		--	end if
		--end for

		for j=1 to 2 do
			if j=1 then
				pat = {
				"mov eax,/var/",
				"push dword ptr [eax]",
				"add dword ptr [esp],/imm/",
				"push $1",
				"mov eax,[esp]",
				"mov ecx,[esp+4]",
				"mov [eax],ecx",
				"add esp,8"}
				r = {
				"mov eax,$1",
				"add dword ptr [eax],$2"
				}
			elsif j=2 then
				pat = {
				"mov eax,/var/",
				"push dword ptr [eax]",
				"mov eax,/imm/",
				"pop ecx",
				"cmp ecx,eax"}
				r = {
				"mov eax,$1",
				"cmp dword ptr [eax],$2"}
			end if
			
			p = 1
			while 1 do
				if p+7>length(subject) then
					exit
				end if
				n = 1
				patvars = {}
				for i=1 to length(pat) do
					if not compare_patterns(subject[p+i-1],pat[i]) then
						n = 0
						exit
					end if
				end for
				if n then
					s = {}
					for i=1 to length(r) do
						s = pattern_append(s,r[i])
					end for
					if p>1 and p+length(pat)<=length(subject) then
						subject = subject[1..p-1] & s & subject[p+length(pat)..length(subject)]
						linesRemoved += length(pat)-length(r)
					end if
				end if
				p += 1
			end while
		end for		
	end if
	
	--patvars = {}
	--? compare_patterns("push dword ptr [var_001]","push /var/")
	--? patvars
	
	return subject
end function


